#include <stdio.h>
#include <stdlib.h>

#include <mpi.h>

#define PRINTMTX 0

int main (int argc, char ** argv) {

  int n, i, j;
  double stime, etime;
  int size, rank, rlen;
  char myname[MPI_MAX_PROCESSOR_NAME]; 

  MPI_Status status;

  MPI_Init (&argc, &argv);

  MPI_Comm_size (MPI_COMM_WORLD, &size);
  
  MPI_Comm_rank (MPI_COMM_WORLD, &rank);

  MPI_Get_processor_name (myname, &rlen);

  fprintf (stdout, "I am proces %3d of %3d my name is %s (%s)\n", 
    rank, size, myname, argv[0]);

  fflush(stdout);

  if (argc != 2) {
    fprintf (stderr, "usage: %s N\n", argv[0]);
    MPI_Finalize ();
    return EXIT_FAILURE;
  } 

  if ( (n = atoi (argv[1])) < 0) {
    fprintf (stderr, "N must be gt 0\n");
    MPI_Finalize ();
    return EXIT_FAILURE;
  }

  if (size == 2) {
    /* primo modo */
    {
      double ** d;

      srand (10);

      if ( (d = (double **) malloc (sizeof (double *) * n)) == NULL) {
        fprintf (stderr, "Memory allocation error\n");
        return EXIT_FAILURE;
      }
      else {
        for (i=0; i<n; i++) {
          if ( (d[i] = (double *) malloc (sizeof (double) * n)) == NULL) {
            fprintf (stderr, "Memory allocation error\n");
            return EXIT_FAILURE;
          }
        }
      }

      if (rank == 0) {
        for (i=0; i<n; i++) 
          for (j=0; j<n; j++)
           d[i][j] = 1.0e0 + (double) (1.0e0 * n * (rand() / 
                 (RAND_MAX + 1.0e0))); 
      }

      if (rank == 0) {
        stime = MPI_Wtime ();
        for (i=0; i<n; i++) 
          MPI_Send (d[i], n, MPI_DOUBLE, 1, i, MPI_COMM_WORLD);
        etime = MPI_Wtime ();
        fprintf (stdout, "Send time 1: %f s\n", etime - stime);
      }
      else {
        stime = MPI_Wtime ();
        for (i=0; i<n; i++)
          MPI_Recv (d[i], n, MPI_DOUBLE, 0, i, MPI_COMM_WORLD, 
              &status);
        etime = MPI_Wtime ();
        fprintf (stdout, "Recv time 1: %f s\n", etime - stime);
      }

#if PRINTMTX
      if (rank == 1) {
        fprintf (stdout, "\n");
        for (i=0; i<n; i++) {
          for (j=0; j<n; j++) {
            fprintf (stdout, "%10.5f ", d[i][j]);
          }
          fprintf (stdout, "\n");
        }

        fflush (stdout);
      }
#endif

      for (i=0; i<n; i++)
        free (d[i]);
      free (d);
    }

    /* secondo */
    {
      double ** d, * temp;

      srand (10);

      if ( (d = (double **) malloc (sizeof (double *) * n)) == NULL) {
        fprintf (stderr, "Memory allocation error\n");
        return EXIT_FAILURE;
      }
      else {
        if ( (temp = (double *) malloc (sizeof (double) * n * n)) == NULL) {
          fprintf (stderr, "Memory allocation error\n");
          return EXIT_FAILURE;
        }
        else {
          for (i=0; i<n; i++) 
            d[i] = &temp[i*n]; 
        }
      }

      if (rank == 0) {
        for (i=0; i<n; i++) 
          for (j=0; j<n; j++)
            d[i][j] = 1.0e0 + (double) (1.0e0 * n * (rand() / (RAND_MAX + 1.0e0))); 
      }

      if (rank == 0) {
        stime = MPI_Wtime ();
        MPI_Send (temp, n*n, MPI_DOUBLE, 1, 1, MPI_COMM_WORLD);
        etime = MPI_Wtime ();
        fprintf (stdout, "Send time 2: %f s\n", etime - stime);
      }
      else {
        stime = MPI_Wtime ();
        MPI_Recv (temp, n*n, MPI_DOUBLE, 0, 1, MPI_COMM_WORLD, 
             &status);
        etime = MPI_Wtime ();
        fprintf (stdout, "Recv time 2: %f s\n", etime - stime);
      }

#if PRINTMTX
      if (rank == 1) {
        fprintf (stdout, "\n");
        for (i=0; i<n; i++) {
          for (j=0; j<n; j++) {
            fprintf (stdout, "%10.5f ", d[i][j]);
          }
          fprintf (stdout, "\n");
        }
      }
#endif

      free (d);
      free (temp);
    }

    /* ancora */
    {
      double * d;

      srand (10);

      if ( (d = (double *) malloc (sizeof (double) * n * n)) == NULL) {
        fprintf (stderr, "Memory allocation error\n");
        return EXIT_FAILURE;
      }

      if (rank == 0) {
        for (i=0; i<n; i++)
          for (j=0; j<n; j++)
            d[(i*n) + j] = 1.0e0 + (double) (1.0e0 * n * (rand() / (RAND_MAX + 1.0e0)));   
      }

      if (rank == 0) {
        stime = MPI_Wtime ();
        MPI_Send (d, n*n, MPI_DOUBLE, 1, 1, MPI_COMM_WORLD);
        etime = MPI_Wtime ();
        fprintf (stdout, "Send time 3: %f s\n", etime - stime);
      }
      else {
        stime = MPI_Wtime ();
        MPI_Recv (d, n*n, MPI_DOUBLE, 0, 1, MPI_COMM_WORLD, 
            &status);
        etime = MPI_Wtime ();
        fprintf (stdout, "Recv time 3: %f s\n", etime - stime);
      }

#if PRINTMTX
      if (rank == 1) {
        fprintf (stdout, "\n");
        for (i=0; i<n; i++) {
          for (j=0; j<n; j++) {
            fprintf (stdout, "%10.5f ", d[(i*n) + j]);
          }
          fprintf (stdout, "\n");
        }
      }
#endif

      free (d);

    }
  }

  MPI_Finalize ();

  return EXIT_SUCCESS;
}
